//## begin module%1.4%.codegen_version preserve=yes
//   Read the documentation to learn more about C++ code generator
//   versioning.
//## end module%1.4%.codegen_version

//## begin module%41F9443502D2.cm preserve=no
//	  %X% %Q% %Z% %W%
//## end module%41F9443502D2.cm

//## begin module%41F9443502D2.cp preserve=no
//## end module%41F9443502D2.cp

//## Module: MarginGrid%41F9443502D2; Package body
//## Subsystem: XSFit::MCMC%41E6AA18032C
//## Source file: /local/data/xerxes1/gordon/headas/Xspec/src/XSFit/MCMC/MarginGrid.cxx

//## begin module%41F9443502D2.additionalIncludes preserve=no
//## end module%41F9443502D2.additionalIncludes

//## begin module%41F9443502D2.includes preserve=yes
#include <XSFit/MCMC/Chain.h>
#include <XSModel/Model/Model.h>
#include <XSUtil/Numerics/ModularCounter.h>
#include <XSstreams.h>
#include <sstream>
//## end module%41F9443502D2.includes

// ChainManager
#include <XSFit/MCMC/ChainManager.h>
// MarginGrid
#include <XSFit/MCMC/MarginGrid.h>


//## begin module%41F9443502D2.declarations preserve=no
//## end module%41F9443502D2.declarations

//## begin module%41F9443502D2.additionalDeclarations preserve=yes
//## end module%41F9443502D2.additionalDeclarations


// Class MarginGrid::MarginInfo 

// Class MarginGrid 

MarginGrid::MarginGrid (ChainManager* chainManager)
  //## begin MarginGrid::MarginGrid%41FA77E901D7.hasinit preserve=yes
  : Grid(),
    m_isSimple(false),
    m_marginInfo(),
    m_chainManager(chainManager),
    m_paramIDs()
  //## end MarginGrid::MarginGrid%41FA77E901D7.hasinit
  //## begin MarginGrid::MarginGrid%41FA77E901D7.initialization preserve=yes
  //## end MarginGrid::MarginGrid%41FA77E901D7.initialization
{
  //## begin MarginGrid::MarginGrid%41FA77E901D7.body preserve=yes
  //## end MarginGrid::MarginGrid%41FA77E901D7.body
}


MarginGrid::~MarginGrid()
{
  //## begin MarginGrid::~MarginGrid%41F944FD021B_dest.body preserve=yes
  //## end MarginGrid::~MarginGrid%41F944FD021B_dest.body
}



//## Other Operations (implementation)
void MarginGrid::doGrid ()
{
  //## begin MarginGrid::doGrid%41F9474C0311.body preserve=yes
   setMarginInfo();
   ProbabilityHist histogram(m_marginInfo.minRanges, m_marginInfo.maxRanges,
                 m_marginInfo.nSteps, m_marginInfo.isLog);
   const ChainManager::ChainContainer& chains = m_chainManager->chains();
   if (!chains.size())
   {
      throw Chain::ChainError("Cannot bin probabilities, no chains currently loaded.");
   }
   const SpecContainer& paramSpecs = getParameter();
   const size_t nPars = paramSpecs.size();
   if (nPars != histogram.nDim())
   {
      throw RedAlert("nParams and Histogram dimensions mismatch.");
   }
   // Bundle user-entered parameter specs into form that
   // Chain can deal with:
   m_paramIDs.resize(nPars);
   for (size_t i=0; i<nPars; ++i)
   {
      string modName = paramSpecs[i]->name;
      if (modName.empty())  modName = Model::DEFAULT();
      m_paramIDs[i].modName = modName;
      m_paramIDs[i].index = paramSpecs[i]->parIndex;
      // ParamID.parName and units are not needed for lookup
      // in Chain.  If Chain finds par based on modName and index,
      // it will fill in the rest of m_paramIDs[i] info.
   }
   
   ChainManager::ChainContainer::const_iterator itChain = chains.begin();
   ChainManager::ChainContainer::const_iterator itEnd = chains.end();
   while (itChain != itEnd)
   {
      std::vector<size_t> found;
      const Chain* chain = itChain->second;
      // this function fills in rest of paramID struct for
      // each requested parameter it finds in Chain.
      chain->findParsInChain(m_paramIDs, found);
      if (found.size() != m_paramIDs.size())
      {
         string msg("Unable to locate all requested parameters in chain file ");
         msg += chain->getFileName();
         throw Chain::ChainError(msg);
      }
      // Check for simple case of requested params
      // matching up 1 to 1 and in order with those 
      // stored in chain file, and treat this as special case
      // which won't need time-consuming makePoint procedure.
      m_isSimple = true;
      if (found.size() == chain->paramIDs().size())
      {
         for (size_t j=0; j<found.size(); ++j)
         {
            if (found[j] != j)
            {
               m_isSimple = false;
               break;
            }
         }
      }
      else
      {
         m_isSimple = false;
      } 
      try
      { 
         chain->openForReadPoints();
         size_t nPoints = chain->length();
         if (m_isSimple)
         {
            std::vector<Real> point;
            for (size_t i=0; i<nPoints; ++i)
            {
               chain->readPoint(point);
               // remove ChiSquare value from point array
               point.pop_back();
               histogram.addToHist(point);
            }
         }
         else
         {
            std::vector<Real> point;
            for (size_t i=0; i<nPoints; ++i)
            {
               chain->readPoint(point);
               makePoint(found, point);
               histogram.addToHist(point);
            }
         }
      }
      catch (...)
      {
         chain->closeFile();
         throw;
      }
      chain->closeFile();
      ++itChain;
   } // end chain loop
   // Convert histogram counts to probabilities.
   long allPoints = histogram.totalAttempts();
   if (!allPoints)
   {
      throw Chain::ChainError("Cannot calculate probabilities, no points found in chains.");
   }
   // NOTE: Up to this point, both Grid::m_grid and ParameterSpec::
   //  parameterValues have been sized by the number of BOUNDARY
   // values.  Since this is a histogram though, we ultimately are
   // concerned with the CENTERS of the bins.  The following function
   // call therefore resizes these arrays.   
   calcBinCenters();
   RealArray& probabilities = grid();
   const std::valarray<long>& counts = histogram.counts();
   size_t gridSz = probabilities.size();
   if (gridSz != counts.size())
   {
      throw RedAlert("Margin grid - histogram size mismatch");
   }
   for (size_t i=0; i<gridSz; ++i)
   {
      probabilities[i] = static_cast<Real>(counts[i]);
   }
   probabilities /= allPoints;
  //## end MarginGrid::doGrid%41F9474C0311.body
}

void MarginGrid::doPoint ()
{
  //## begin MarginGrid::doPoint%41F9476F026F.body preserve=yes
  //## end MarginGrid::doPoint%41F9476F026F.body
}

void MarginGrid::report (bool title) const
{
  //## begin MarginGrid::report%41F947EC01C1.body preserve=yes
   using namespace std;
   const RealArray& probs = getGrid();
   const size_t probSz = probs.size();
   if (probSz)
   {
      ios_base::fmtflags saveFlags(tcout.flags());
      streamsize savePrecision = tcout.precision();
      size_t nPars = m_marginInfo.minRanges.size();
      const int precision = 6;
      const int fwidth = precision + 8;
      tcout.precision(precision);  
      tcout << showpoint; 
      tcout << left << setw(fwidth) << "Probability" << right;
      for (size_t i=0; i<nPars; ++i)
      {
         tcout << setw(fwidth) << "Mod param";
      }
      tcout << endl << setw(fwidth) << " ";
      const SpecContainer& paramSpecs = getParameter();
      for (size_t i=0; i<nPars; ++i)
      {
         ostringstream tmpSS;
         if (paramSpecs[i]->name.length())
         {
            tmpSS << paramSpecs[i]->name << ":";
         }
         tmpSS << paramSpecs[i]->parIndex;
         tcout << setw(fwidth) << tmpSS.str();
      }
      tcout << endl;
      Numerics::ModularCounter idxCounter(m_marginInfo.nSteps);
      const IntegerArray& currCounts = idxCounter.counter();
      ios_base& (*formatType) (ios_base& ) = fixed;
      for (size_t i=0; i<probSz; ++i)
      {
         formatType = (probs[i]>.0 && abs(log10(probs[i])) >= 2.0)
                 ? scientific : fixed; 
         tcout << setw(fwidth) << formatType << probs[i];
         for (size_t j=0; j<nPars; ++j)
         {
            ParameterSpec* par = paramSpecs[j];
            Real val = par->parameterValues[currCounts[j]];
            formatType = (val>.0 && abs(log10(val)) >= 2.0)
                 ? scientific : fixed; 
            tcout << setw(fwidth) << formatType << val;
         }
         ++idxCounter;
         tcout << endl;
      }
      tcout.precision(savePrecision);
      tcout.flags(saveFlags);
   }
  //## end MarginGrid::report%41F947EC01C1.body
}

void MarginGrid::setMarginInfo ()
{
  //## begin MarginGrid::setMarginInfo%41FA760100F3.body preserve=yes
   const SpecContainer& info = getParameter();
   size_t sz = info.size();
   m_marginInfo.minRanges.resize(sz);
   m_marginInfo.maxRanges.resize(sz);
   m_marginInfo.nSteps.resize(sz);
   m_marginInfo.isLog.resize(sz);
   for (size_t i=0; i<sz; ++i)
   {
      m_marginInfo.minRanges[i] = info[i]->lowRange;
      m_marginInfo.maxRanges[i] = info[i]->highRange;
      m_marginInfo.nSteps[i] = info[i]->intervals;
      m_marginInfo.isLog[i] = info[i]->log;
   }
  //## end MarginGrid::setMarginInfo%41FA760100F3.body
}

void MarginGrid::makePoint (const std::vector<size_t>& indices, std::vector<Real>& point)
{
  //## begin MarginGrid::makePoint%41FA761701F4.body preserve=yes
    size_t sz = indices.size();
   std::vector<Real> tmp;
   // CAUTION!: This function ASSUMES all indices have been verified
   // to fall within size of point vector.  In effort to speed 
   // things up a bit, it will not check for them here.
   for (size_t i=0; i<sz; ++i)
   {
      tmp.push_back(point[indices[i]]);
   }
   std::swap(point, tmp);
  //## end MarginGrid::makePoint%41FA761701F4.body
}

void MarginGrid::calcBinCenters ()
{
  //## begin MarginGrid::calcBinCenters%41FE6B4A010B.body preserve=yes
   // Replace bin boundary values in ParameterSpec::parameterValue
   // arrays with their center values, and resize arrays.
   SpecContainer& parameters = parameter();
   size_t totSize = 1;
   for (size_t i=0; i<parameters.size(); ++i)
   {
      ParameterSpec* par = parameters[i];
      std::vector<Real>& parVals = par->parameterValues;
      if (par->log)
      {
         size_t nBins = parVals.size()-1;
         for (size_t j=0; j<nBins; ++j)
         {
            parVals[j] = sqrt(parVals[j]*parVals[j+1]);
         }
      }
      else
      {
         size_t nBins = parVals.size()-1;
         for (size_t j=0; j<nBins; ++j)
         {
            parVals[j] = (parVals[j] + parVals[j+1])/2.0;
         }
      }
      parVals.pop_back();
      totSize *= parVals.size();      
   }
   grid().resize(totSize,.0);
   
  //## end MarginGrid::calcBinCenters%41FE6B4A010B.body
}

// Additional Declarations
  //## begin MarginGrid%41F944FD021B.declarations preserve=yes
  //## end MarginGrid%41F944FD021B.declarations

//## begin module%41F9443502D2.epilog preserve=yes
//## end module%41F9443502D2.epilog
